/*
  Copyright (c) 2006 XenSource, Inc.  All rights reserved.
*/

#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>

#include <libxml/xmlsave.h>
#include <libxml/xmlstring.h>
#include <libxml/tree.h>

#include "base.h"

#include "xen_common.h"
#include "xen_internal.h"


typedef struct
{
    int size;
    void *contents;
} arbitrary_set;


static char *
make_body(const char *method_name, abstract_value params[], int param_count);

static void
parse_result(xen_session *session, const char *result,
             abstract_type result_type, void *value);

static void
add_param(xmlNodePtr params_node, const char *name, const char *value);

static void
call_raw(xen_session *s, const char *method_name,
         abstract_value params[], int param_count, abstract_type result_type,
         void *value);


xen_session *
xen_session_login_with_password(xen_call_func call_func, void *handle,
                                const char *uname, const char *pwd)
{
    abstract_value params[] =
        {
            { .type.typename = STRING,
              .u.string_val = uname },
            { .type.typename = STRING,
              .u.string_val = pwd }
        };
    abstract_type result_type = { .typename = STRING };

    xen_session *session = malloc(sizeof(xen_session));
    session->call_func = call_func;
    session->handle = handle;
    session->session_id = NULL;
    session->ok = true;
    session->error_description = NULL;
    session->error_description_count = 0;

    call_raw(session, "Session.login_with_password", params,
             sizeof(params) / sizeof(params[0]),
             result_type, &session->session_id);

    return session;
}


void
xen_session_logout(xen_session *session)
{
    abstract_value params[] =
        {
            { .type.typename = STRING,
              .u.string_val = session->session_id }
        };
    abstract_type result_type = { .typename = VOID };

    xen_call_(session, "Session.logout", params,
              sizeof(params) / sizeof(params[0]), result_type, NULL);

    if (session->error_description != NULL)
    {
        for (int i = 0; i < session->error_description_count; i++)
        {
            free(session->error_description[i]);
        }
        free(session->error_description);
    }

    free(session);
}


/**
 * @param value A pointer to the correct location as per the given
 * result_type.  Will be populated if the call succeeds.  In that case, and if
 * value is a char **, the char * itself must be freed by the caller.
 */
void
xen_call_(xen_session *s, const char *method_name,
          abstract_value params[], int param_count, abstract_type result_type,
          void *value)
{
    if (!s->ok)
    {
        return;
    }

    abstract_value *full_params =
        malloc(sizeof(abstract_value) * (param_count + 1));

    full_params[0].type.typename = STRING;
    full_params[0].u.string_val = s->session_id;

    memcpy(full_params + 1, params, param_count * sizeof(abstract_value));

    call_raw(s, method_name, full_params, param_count + 1, result_type,
             value);
}


static void
call_raw(xen_session *s, const char *method_name,
         abstract_value params[], int param_count, abstract_type result_type,
         void *value)
{
    char *body = make_body(method_name, params, param_count);
    char *result = s->call_func(body, strlen(body), s->handle);
    free(body);
    parse_result(s, result, result_type, value);
    free(result);
}


static void fault(xen_session *session, xmlrpc_env *env)
{
    char ** strings = malloc(3 * sizeof(char *));

    strings[0] = xen_strdup_("SERVER_FAULT");
    strings[1] = malloc(20);
    snprintf(strings[1], 20, "%d", env->fault_code);
    strings[2] = xen_strdup_(env->fault_string);

    session->ok = false;
    session->error_description = strings;
    session->error_description_count = 3;

    xmlrpc_env_clean(env);
}


static void parse_into(xmlrpc_env *env, xmlrpc_value *value_val,
                       abstract_type result_type, void *value, int slot)
{
    switch (result_type.typename)
    {
    case STRING:
        xmlrpc_read_string(env, value_val, ((const char **)value) + slot);
        break;

    case INT:
        xmlrpc_read_int(env, value_val, ((int *)value) + slot);
        break;

    case FLOAT:
        xmlrpc_read_double(env, value_val, ((double *)value) + slot);
        break;

    case BOOL:
        xmlrpc_read_bool(env, value_val, ((int *)value) + slot);
        break;

    case VOID:
        break;

    case SET:
    {
        int n = xmlrpc_array_size(env, value_val);
        void *contents = malloc(sizeof(void *) * n);
        for (int i = 0; i < n; i++)
        {
            xmlrpc_value * member_val;
            xmlrpc_array_read_item(env, value_val, i, &member_val);

            parse_into(env, member_val, *result_type.child1, contents, i);

            xmlrpc_DECREF(member_val);
        }
        arbitrary_set *set = ((arbitrary_set *)value) + slot;
        set->size = n;
        set->contents = contents;
    }
    break;

    default:
        assert(false);
    }
}


/**
 * Parameters as for call() above.
 */
static void parse_result(xen_session *session, const char *result,
                         abstract_type result_type, void *value)
{
    xmlrpc_env env;
    xmlrpc_env_init(&env);

    xmlrpc_value *parsed_result =
        xmlrpc_parse_response(&env, result, strlen(result));

    if (env.fault_occurred)
    {
        return fault(session, &env);
    }

    xmlrpc_value *status_val;
    const char *status_str;

    xmlrpc_struct_find_value(&env, parsed_result, "Status", &status_val);
    xmlrpc_read_string(&env, status_val, &status_str);

    xmlrpc_DECREF(status_val);

    if (env.fault_occurred)
    {
        return fault(session, &env);
    }

    bool success = (0 == strcmp(status_str, "Success"));

    free((char *)status_str);

    if (success)
    {
        xmlrpc_value *value_val;
        xmlrpc_struct_find_value(&env, parsed_result, "Value", &value_val);

        parse_into(&env, value_val, result_type, value, 0);

        xmlrpc_DECREF(value_val);
    }
    else
    {
        xmlrpc_value *error_desc_val;
        xmlrpc_struct_find_value(&env, parsed_result, "ErrorDescription",
                                 &error_desc_val);

        int n = xmlrpc_array_size(&env, error_desc_val);

        session->ok = false;
        session->error_description = malloc(n * sizeof(char *));
        session->error_description_count = n;

        for (int i = 0; i < n; i++)
        {
            xmlrpc_value * error_str;
            xmlrpc_array_read_item(&env, error_desc_val, i, &error_str);
            xmlrpc_read_string(&env, error_str,
                               (const char **)session->error_description + i);
            xmlrpc_DECREF(error_str);
        }

        xmlrpc_DECREF(error_desc_val);
    }

    xmlrpc_DECREF(parsed_result);
    xmlrpc_env_clean(&env);
}


static char *
make_body(const char *method_name, abstract_value params[], int param_count)
{
    char buf[20];

    xmlDocPtr doc = xmlNewDoc(BAD_CAST "1.0");
    xmlNodePtr methodCall = xmlNewNode(NULL, BAD_CAST "methodCall");
    xmlDocSetRootElement(doc, methodCall);

    xmlNewChild(methodCall, NULL, BAD_CAST "methodName",
                BAD_CAST method_name);

    xmlNodePtr params_node =
        xmlNewChild(methodCall, NULL, BAD_CAST "params", NULL);

    for (int p = 0; p < param_count; p++)
    {
        abstract_value v = params[p];
        switch (v.type.typename)
        {
        case STRING:
            add_param(params_node, "string", v.u.string_val);
            break;

        case INT:
            snprintf(buf, sizeof(buf), "%"PRId64, v.u.int_val);
            add_param(params_node, "string", buf);
            break;

        case FLOAT:
            snprintf(buf, sizeof(buf), "%lf", v.u.double_val);
            add_param(params_node, "double", buf);
            break;

        case BOOL:
            add_param(params_node, "boolean", v.u.bool_val ? "1" : "0");
            break;
            
        case VOID:
            add_param(params_node, "string", "");
            break;
            
        default:
            assert(false);
        }
    }

    xmlBufferPtr buffer = xmlBufferCreate();
    xmlSaveCtxtPtr save_ctxt =
        xmlSaveToBuffer(buffer, NULL, XML_SAVE_NO_XHTML);

    if (xmlSaveDoc(save_ctxt, doc) == -1)
    {
        return NULL;
    }

    xmlFreeDoc(doc);
    xmlSaveClose(save_ctxt);
    xmlChar *content = xmlStrdup(xmlBufferContent(buffer));
    xmlBufferFree(buffer);
    return (char *)content;
}


static void
add_param(xmlNodePtr params_node, const char *name, const char *value)
{
  xmlNodePtr param_node =
    xmlNewChild(params_node, NULL, BAD_CAST "name",  NULL);
  xmlNodePtr value_node =
    xmlNewChild(param_node,  NULL, BAD_CAST "value", NULL);

  xmlNewChild(value_node, NULL, BAD_CAST name, BAD_CAST value);
}


char *
xen_strdup_(char *in)
{
    char *result = malloc(strlen(in) + 1);
    strcpy(result, in);
    return result;
}
